import os
import numpy as np

import torch
import torch.nn as nn
import torch.backends.cudnn as cudnn
import densenet
import torchvision.transforms as transforms
from torch.utils.data import DataLoader
from sklearn.metrics import roc_auc_score
import torch.nn.functional as F
import torchvision

class DenseNet121(nn.Module):

    def __init__(self, classCount, isTrained):
	
        super(DenseNet121, self).__init__()
		
        self.densenet121 = torchvision.models.densenet121(pretrained=isTrained)

        kernelCount = self.densenet121.classifier.in_features
		
        self.densenet121.classifier = nn.Sequential(nn.Linear(kernelCount, classCount), nn.Sigmoid())

    def forward(self, x):
        x = self.densenet121(x)
        return x


# class model_densenet121(nn.Module):
#
#     def __init__(self, classCount, isTrained):
#         super(model_densenet121, self).__init__()
#
#         self.densenet121 = torchvision.models.densenet121(pretrained=isTrained)
#
#         kernelCount = self.densenet121.classifier.in_features
#
#         self.densenet121.classifier = nn.Sequential(nn.Linear(kernelCount, classCount), nn.Sigmoid())
#
#     def forward(self, x):
#         x = self.densenet121(x)
#         return x

class DenseNet169(nn.Module):
    
    def __init__(self, classCount, isTrained):
        
        super(DenseNet169, self).__init__()
        
        self.densenet169 = torchvision.models.densenet169(pretrained=isTrained)
        
        kernelCount = self.densenet169.classifier.in_features
        
        self.densenet169.classifier = nn.Sequential(nn.Linear(kernelCount, classCount), nn.Sigmoid())
        
    def forward (self, x):
        x = self.densenet169(x)
        return x
    
class DenseNet201(nn.Module):
    
    def __init__ (self, classCount, isTrained):
        
        super(DenseNet201, self).__init__()
        
        self.densenet201 = torchvision.models.densenet201(pretrained=isTrained)
        
        kernelCount = self.densenet201.classifier.in_features
        
        self.densenet201.classifier = nn.Sequential(nn.Linear(kernelCount, classCount), nn.Sigmoid())
        
    def forward (self, x):
        x = self.densenet201(x)
        return x

class ms_cam(nn.Module):
    def __init__(self,in_c,r,BatchNorm2d=nn.BatchNorm2d):
        super(ms_cam,self).__init__()
        self.branch1= nn.Sequential(nn.AdaptiveAvgPool2d((1,1)),
                                    nn.Conv2d(in_c,in_c//r,kernel_size=1,stride=1,bias=False),
                                    BatchNorm2d(in_c//r),
                                    nn.ReLU(),
                                    nn.Conv2d(in_c // r, in_c, kernel_size=1, stride=1, bias=False),
                                    BatchNorm2d(in_c),
                                    )

        self.branch2 = nn.Sequential(
            nn.Conv2d(in_c,in_c//r,kernel_size=1,stride=1,bias=False),
            BatchNorm2d(in_c // r),
            nn.ReLU(),
            nn.Conv2d(in_c//r, in_c, kernel_size=1, stride=1, bias=False),
            BatchNorm2d(in_c)
        )
    def forward(self,x):

        x1 = self.branch1(x)
        x2 = self.branch2(x)
        x3 = x1 + x2
        x3 = torch.sigmoid(x3)

        return x3

class iAFF(nn.Module):
    def __init__(self,in_c,r,BatchNorm2d=nn.BatchNorm2d):
        super(iAFF,self).__init__()
        self.ms_cam1 = ms_cam(in_c,r,BatchNorm2d=BatchNorm2d)
        self.ms_cam2 = ms_cam(in_c, r,BatchNorm2d=BatchNorm2d)

    def forward(self,x_1,x_2):
        x1_sum= x_1 + x_2
        amap1 = self.ms_cam1(x1_sum)

        x_1_a = amap1 * x_1
        x_2_a = (1 - amap1) * x_2

        x2_sum = x_1_a + x_2_a
        amap2 = self.ms_cam2(x2_sum)

        x_1 = amap2 * x_1
        x_2 = (1 - amap2) * x_2

        x = x_1 + x_2
        return x

class ms_fusion(nn.Module):
    def __init__(self,in_c,out_c,r,BatchNorm2d=nn.BatchNorm2d):
        super(ms_fusion,self).__init__()
        self.iaff = iAFF(out_c,r,BatchNorm2d=BatchNorm2d)
        self.down_sample = nn.Sequential(
            nn.Conv2d(in_c,out_c,kernel_size=3,padding=1,stride=2,bias=False),
            nn.BatchNorm2d(out_c)
        )

    def forward(self,x_1,x_2):
        x_1 = self.down_sample(x_1)
        x = self.iaff(x_1,x_2)
        return x

class model_densenet121(nn.Module):
    def __init__(self, classCount, isTrained,BatchNorm2d=nn.BatchNorm2d):
        super(model_densenet121, self).__init__()

        self.backbone1 = densenet.densenet169(pretrained=isTrained)  #1024
        self.backbone2 = densenet.densenet169(pretrained=isTrained)
        # torch.Size([2, 256, 56, 56])
        # torch.Size([2, 512, 28, 28])
        # torch.Size([2, 1024, 14, 14])
        # torch.Size([2, 1024, 7, 7])
        self.ms_fusion1 = ms_fusion(256,512,8,BatchNorm2d=BatchNorm2d)
        self.ms_fusion2 = ms_fusion(512,1280,8,BatchNorm2d=BatchNorm2d)
        self.ms_fusion3 = ms_fusion(1280,1664,8,BatchNorm2d=BatchNorm2d)

        self.pool = nn.AdaptiveAvgPool2d((1,1))
        #self.fc = nn.Linear(2048, classCount)
        self.fc1 = nn.Linear(1664*2,classCount)
        self.fc2 = nn.Linear(1664, classCount)

    def forward(self, x):
        x1,x2,x3,x4 = self.backbone1._forward(x)
        _,_,_, x4_2 = self.backbone2._forward(x)
        x_f = self.ms_fusion1(x1,x2)
        x_f = self.ms_fusion2(x_f,x3)
        x_f = self.ms_fusion3(x_f,x4)
        feat1 = self.pool(x4)
        feat1 = feat1.squeeze(-1).squeeze(-1)
        feat2 = self.pool(x_f)
        feat2 = feat2.squeeze(-1).squeeze(-1)
        feat3 = self.pool(x4_2)
        feat3 = feat3.squeeze(-1).squeeze(-1)
        y1 = self.fc1(torch.cat([feat1,feat2],dim=1))
        y2 = self.fc2(feat3)
        y = (y2 + y1)/2.0
        return torch.sigmoid(y)
        
class model_densenet121_att(nn.Module):
    def __init__(self, classCount, isTrained,BatchNorm2d=nn.BatchNorm2d):
        super(model_densenet121_att, self).__init__()

        self.backbone1 = densenet.densenet169(pretrained=isTrained)  #1024
        self.backbone2 = densenet.densenet169(pretrained=isTrained)
        # torch.Size([2, 256, 56, 56])
        # torch.Size([2, 512, 28, 28])
        # torch.Size([2, 1024, 14, 14])
        # torch.Size([2, 1024, 7, 7])
        self.ms_fusion1 = ms_fusion(256,512,8,BatchNorm2d=BatchNorm2d)
        self.ms_fusion2 = ms_fusion(512,1280,8,BatchNorm2d=BatchNorm2d)
        self.ms_fusion3 = ms_fusion(1280,1664,8,BatchNorm2d=BatchNorm2d)

        self.pool = nn.AdaptiveAvgPool2d((1,1))
        #self.fc = nn.Linear(2048, classCount)
        self.fc1 = nn.Linear(1664*2,classCount)
        self.fc2 = nn.Linear(1664, classCount)
    def getacm(self,y,classifier):
        b, c, h, w = y.size()
        y = y.permute([0, 2, 3, 1]).contiguous().view(-1, c)
        y = classifier(y)
        y = y.view(b, h, w, -1).permute([0, 3, 1, 2])
        return y
    def forward(self, x):
        x_flip = torch.flip(x,[3])
        x1,x2,x3,x4 = self.backbone1._forward(x)
        _,_,_, x4_flip = self.backbone2._forward(x_flip)
        x4_2 = torch.flip(x4_flip,[3])
        x_f = self.ms_fusion1(x1,x2)
        x_f = self.ms_fusion2(x_f,x3)
        x_f = self.ms_fusion3(x_f,x4)
        feat1 = torch.cat([x4,x_f],dim=1)
        cam1 = self.getacm(feat1,self.fc1)
        feat1 = self.pool(feat1)
        feat1 = feat1.squeeze(-1).squeeze(-1)
        cam2 = self.getacm(x4_2, self.fc2)
        feat2 = self.pool(x4_2)
        feat2 = feat2.squeeze(-1).squeeze(-1)

        y1 = self.fc1(feat1)
        y2 = self.fc2(feat2)
        y = (y2 + y1)/2.0

        return torch.sigmoid(y),cam1,cam2
        