import os
import torch
import torch.nn as nn
import torch.nn.functional as F
from sklearn.metrics import confusion_matrix
from tqdm import tqdm
from tensorboardX import SummaryWriter
import numpy as np
import time
from sklearn.metrics import roc_auc_score

class AverageMeter(object):
    """Computes and stores the average and current value"""
    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

def intersectionAndUnion(output, target, K, ignore_index=255):
    # 'K' classes, output and target sizes are N or N * L or N * H * W, each value in range 0 to K - 1.
    assert (output.ndim in [1, 2, 3])
    assert output.shape == target.shape

    output = output.reshape(output.size).copy()
    target = target.reshape(target.size)

    output[np.where(target == ignore_index)[0]] = 255

    intersection = output[np.where(output == target)[0]]
    area_intersection, _ = np.histogram(intersection, bins=np.arange(K+1))
    area_output, _ = np.histogram(output, bins=np.arange(K+1))
    area_target, _ = np.histogram(target, bins=np.arange(K+1))
    area_union = area_output + area_target - area_intersection

    return area_intersection, area_union, area_target

class Trainer(object):

    def __init__(self, model, optimizer, loss_f, epoch_max=100 ,save_dir=None, save_freq=1):
        self.model = model
        device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        if torch.cuda.device_count() > 1:
            print("Let's use", torch.cuda.device_count(), "GPUs!")
            # dim = 0 [30, xxx] -> [10, ...], [10, ...], [10, ...] on 3 GPUs
            self.model = torch.nn.DataParallel(self.model,device_ids=range(torch.cuda.device_count()))
        self.model.to(device)
        self.model.load_state_dict(torch.load("./ model_50.pkl")['weight'])
        self.optimizer = optimizer
        self.loss_f = loss_f

        self.save_dir = save_dir
        self.save_freq = save_freq
        self.writer = SummaryWriter()
        self.epoch_max =epoch_max


    def _iteration(self, data_loader, ep ,is_train=True):
        if is_train:
            batch_time = AverageMeter()
            data_time = AverageMeter()
            main_loss_meter = AverageMeter()
            loss_meter = AverageMeter()
            loss_2_meter = AverageMeter()
            intersection_meter = AverageMeter()
            union_meter = AverageMeter()
            target_meter = AverageMeter()

            self.model.train()
            # self.model.cuda()
            end = time.time()
            weight = torch.Tensor([ 2, 1, 1, 1, 1, 1,5,2,2,2,2,2,2,5]).cuda().float()
            criterion = nn.BCEWithLogitsLoss(weight=weight)
            for index, (input,label) in enumerate(data_loader):
                input = input.cuda()
                target = label.cuda()
                # print('target.size:',target.size())

                data_time.update(time.time() - end)

                output = self.model(input)
                # b= am1.size()
                # am1 = am1.view(b, -1)
                # am1 = am1 / am1.max(dim=1)[0].unsqueeze(-1)
                # am2 = am2.view(b, -1)
                # am2 = am2 / am2.max(dim=1)[0].unsqueeze(-1)
                # loss_aux = torch.norm(am1 - am2)
                main_loss = criterion(output,target)

                loss = main_loss

                self.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()

                # output = torch.sigmoid(output)
                # output = output.cpu().data.numpy()
                # output[output > 0.5] = 1
                # output[output <= 0.5] = 0
                # target = target.detach().cpu().numpy()
                # intersection, union, target = intersectionAndUnion(output, target, 15, 255)
                # intersection_meter.update(intersection)
                # union_meter.update(union)
                # target_meter.update(target)
                # accuracy = sum(intersection_meter.val) / (sum(target_meter.val) + 1e-10)
                loss_meter.update(float(loss.data), input.size(0))
                loss_2_meter.update(float(loss.data),input.size(0))
                main_loss_meter.update(float(main_loss.data),input.size(0))

                batch_time.update(time.time() - end)


                end = time.time()
                current_iter = (ep - 1) * len(data_loader) + index + 1
                max_iter = self.epoch_max * len(data_loader)
                remain_iter = max_iter - current_iter
                remain_time = remain_iter * batch_time.avg
                t_m, t_s = divmod(remain_time, 60)
                t_h, t_m = divmod(t_m, 60)
                remain_time = '{:02d}:{:02d}:{:02d}'.format(int(t_h), int(t_m), int(t_s))
                if (index + 1) % 10 == 0 :
                    print('Epoch: [{}/{}][{}/{}] '
                          'Batch {batch_time.val:.3f} ({batch_time.avg:.3f}) '
                          'Remain {remain_time} '
                          'Loss {loss_meter:.4f} '
                          'aux{loss_aux:.4f}'
                          'main_loss{main_loss:.4f}'
                          'Accuracy {accuracy:.4f}.'.format(ep, self.epoch_max, index + 1, len(data_loader),
                                                            remain_time = remain_time,
                                                            batch_time=batch_time,
                                                            loss_meter=loss_meter.val,
                                                            main_loss = main_loss_meter.val,
                                                            loss_aux = loss_2_meter.val,
                                                            accuracy=0))
            # accuracy_class = intersection_meter.sum / (target_meter.sum + 1e-10)
            #
            # mAcc = np.mean(accuracy_class)
            # allAcc = sum(intersection_meter.sum) / (sum(target_meter.sum) + 1e-10)
            #
            # print('[train]:epoch [{}/{}]: mAcc/allAcc: {:.4f}/{:.4f}'.format(ep, 100,
            #                                                                              mAcc, allAcc))
        else:
            batch_time = AverageMeter()
            data_time = AverageMeter()
            main_loss_meter = AverageMeter()
            loss_meter = AverageMeter()
            intersection_meter = AverageMeter()
            union_meter = AverageMeter()
            target_meter = AverageMeter()

            self.model.eval()
            # self.model.cuda()

            end = time.time()
            targets = []
            preds=[]
            for index, (input, label) in enumerate(data_loader):
                input = input.cuda()
                # print('input size:',input.size())
                target = label
                data_time.update(time.time() - end)

                output= self.model(input)


                output = torch.sigmoid(output)
                output = output.cpu().data.numpy()
                preds.append(output.copy())
                output[output>0.5]=1
                output[output<=0.5]=0
                target = target.numpy()
                targets.append(target)

                intersection, union, target = intersectionAndUnion(output, target, 2, 255)
                intersection_meter.update(intersection)
                union_meter.update(union)
                target_meter.update(target)
                accuracy = sum(intersection_meter.val) / (sum(target_meter.val) + 1e-10)

                batch_time.update(time.time() - end)
                end = time.time()

                if (index + 1) % 10 == 0:
                    print('Epoch: [{}/{}][{}/{}] '
                          'Batch {batch_time.val:.3f} ({batch_time.avg:.3f}) '
                          'Loss {loss_meter:.4f} '
                          'Accuracy {accuracy:.4f}.'.format(ep, 100, index + 1, len(data_loader),
                                                            batch_time=batch_time,
                                                            loss_meter=loss_meter.val,
                                                            accuracy=accuracy))
            accuracy_class = intersection_meter.sum / (target_meter.sum + 1e-10)
            print(accuracy_class)

            mAcc = np.mean(accuracy_class)
            allAcc = sum(intersection_meter.sum) / (sum(target_meter.sum) + 1e-10)

            print('[val]:epoch [{}/{}]: mAcc/allAcc: {:.4f}/{:.4f}'.format(ep, 100,mAcc, allAcc))

            targets = np.concatenate(targets,axis=0)
            preds = np.concatenate(preds, axis=0)
            print(targets.shape,preds.shape)
            aucs=[]
            for _c in range(preds.shape[-1]):
                aucs.append(roc_auc_score(targets[:,_c],preds[:,_c]))
            print('AUCs:',aucs)
            print('mAUC:',np.mean(aucs))
        return

    def train(self, data_loader,ep):

        loss = self._iteration(data_loader,ep)

    def test(self, data_loader,ep):
        self.model.eval()
        with torch.no_grad():
            loss = self._iteration(data_loader,ep,is_train=False)

    def loop(self, epochs, train_data, test_data, scheduler=None):
        # self.test(test_data, 0)
        for ep in range(1, epochs + 1):
            if scheduler is not None:
                scheduler.step()
            print("epochs: {}".format(ep))
            self.train(train_data,ep)
            if (ep % self.save_freq == 0):
                self.test(test_data, ep)
                self.save(ep)

    def save(self, epoch, **kwargs):
        model_out_path = self.save_dir
        state = {"epoch": epoch, "weight": self.model.state_dict()}
        if not os.path.exists(model_out_path):
            os.makedirs(model_out_path)
        torch.save(state, model_out_path + '/ model_{epoch}.pkl'.format(epoch=epoch))